//----------------------------------------------------------------------------------
// aviutl  GPUJX^tB^
// Copyright (c) 2007  N099
//----------------------------------------------------------------------------------
#include <windows.h>

#include <new>
#include <list>

#include <d3d9.h>
#include <d3dx9.h>
#include <dxerr.h>

#include <strsafe.h>

#include "resource.h"
#include "filter.h"

#include "util.h"
#include "thrproc.h"
#include "effect_reader.h"
#include "gpufilter.h"

//ǂݍލől
static const int TECH_MAX = 32;
static const int TREF_MAX = 32;
static const int TXRT_MAX = 32;
static const int KERN_MAX = 32;

//---------------------------------------------------------------------
// struct
//---------------------------------------------------------------------
struct PPVERT
{
	float x, y, z, rhw;
	float tu, tv;		// Texcoord

	const static D3DVERTEXELEMENT9 Decl[3];
};

const D3DVERTEXELEMENT9 PPVERT::Decl[3] =
{
	{ 0, 0,  D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITIONT, 0 },
	{ 0, 16, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,  0 },
	D3DDECL_END()
};

typedef std::pair<int, IDirect3DTexture9*> Tex_Pair;

//---------------------------------------------------------------------
// public functions
//---------------------------------------------------------------------

//RXgN^
GpuFilter::GpuFilter(EffectReader &er) :
	m_pD3D(NULL), m_pd3dDevice(NULL),
	m_pTexSys(NULL), m_pVert(NULL), m_pVertDec(NULL),
	m_pEffect(NULL), m_pEffectReader(er),
	m_TxCur(true), m_pCpCvt(NULL)
{
	m_hTxTmp = m_hFrame = m_hFrameNum = m_hSrcSize = m_hPixScale = NULL;
	m_tech = 0;
	m_width = m_height = m_wMax = -1;
	m_frame = m_frame_num = -1;
	ZeroMemory(m_frame_set, sizeof(m_frame_set));

	m_pixFormat = Unknown;
	ZeroMemory(m_pixFormatAvailable, sizeof(m_pixFormatAvailable));
	ZeroMemory(&m_caps, sizeof(m_caps));
}

//fXgN^
GpuFilter::~GpuFilter()
{
	//\[X̊J
	ReleaseResources();
	SAFE_RELEASE(m_pEffect);
	m_tproc.Shutdown();

	//AddRefĂ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);
}

//IuWFNg̏
bool GpuFilter::Init(FILTER *fp, int dev, bool pp)
{
	HRESULT hr;

	//Ŝ
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);

	//init D3D    
	m_pD3D = Direct3DCreate9(D3D_SDK_VERSION);
	if(NULL == m_pD3D) {
		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_FAIL, TEXT("Direct3DCreate9()s܂"));
		return false;
	}

	//adapter selection
	UINT adapter = (dev < 0) ? D3DADAPTER_DEFAULT : dev;
	D3DDEVTYPE type = (dev < 0) ? D3DDEVTYPE_REF : D3DDEVTYPE_HAL;
	if(adapter >= m_pD3D->GetAdapterCount()){
		adapter = D3DADAPTER_DEFAULT;
	}

	//create device
	D3DPRESENT_PARAMETERS d3dpp; 

	ZeroMemory(&d3dpp, sizeof(d3dpp));
	d3dpp.Windowed = TRUE;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;

	hr = m_pD3D->CreateDevice(
		adapter, type, fp->hwnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING|D3DCREATE_NOWINDOWCHANGES|D3DCREATE_MULTITHREADED, 
		&d3dpp, &m_pd3dDevice );
	if(FAILED(hr)) {
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DfoCX̍쐬Ɏs܂"));
		SAFE_RELEASE(m_pd3dDevice);
		SAFE_RELEASE(m_pD3D);
		return false;
	}

	//foCX̃`FbN
	if(FAILED(CheckDeviceCap())){
		goto err;
	}

	//XbhpiL̂݁j
	if(pp){
		if(m_tproc.Init(AsyncProc, this)){
			hr = E_FAIL;
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("XbhɎs܂"));
			goto err;
		}
	}

	//GtFNg̍쐬
	hr = CreateEffect();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("GtFNg̍쐬Ɏs܂"));
		goto err;
	}

	return true;

err:
	SAFE_RELEASE(m_pd3dDevice);
	SAFE_RELEASE(m_pD3D);

	return false;
}

//e탊\[X̊m
bool GpuFilter::AllocResources(int w, int h, int w_max)
{
	//̃`FbN
	if(!m_pEffect) return false;

	//sNZTCỸZbg
	if(w<0 || (DWORD)w>m_caps.MaxTextureWidth 
		|| h<0 || (DWORD)h>m_caps.MaxTextureHeight
		|| w_max<0 || w_max<w)
	{
		return false;
	}else{
		m_width = w;
		m_height = h;
		m_wMax = w_max;
	}

	//Rs[p֐擾
	if(!CpCvt::GetCpConverter(m_pCpCvt, PicFmtToCpCvtFmt(m_pixFormat))){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_NOINTERFACE,
			TEXT("摜Rs[ɂMMXKvł"));
		return false;
	}

	//ꉞJĂ
	ReleaseResources();

	//VXeɍ쐬郊\[X
	if(FAILED(CreateSysMemResources())){
		return false;
	}

	//GPUɍ쐬郊\[X
	if(FAILED(CreateDefaultResources())){
		return false;
	}

	//GtFNgp[^̃Zbg
	if(FAILED(SetEffectResources())){
		return false;
	}

	return true;
}

//foCXԂ̃eXg
bool GpuFilter::TestDeviceState(bool *realloc)
{
	HRESULT hr;

	if(!m_pd3dDevice){
		return false;
	}

	// D3D9 device ̏ԂeXg
	hr = m_pd3dDevice->TestCooperativeLevel();
	switch(hr)
	{
		case D3D_OK:
			break;
		case D3DERR_DEVICELOST:
#ifdef _DEBUG
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
					TEXT("foCXXg܂Bs܂B"));
#endif
			return false;
		case D3DERR_DEVICENOTRESET:
			hr = Reset();
			if(FAILED(hr)){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
						TEXT("foCXZbgɎs܂"));
				return false;
			}else{
				*realloc = true;
			}
			break;
		default :
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
					TEXT("TestCooperativeLevel()G[Ԃ܂"));
			return false;
	}

	return true;
}

//e탊\[X̊J
bool GpuFilter::ReleaseResources()
{	
	//\[X̊J
	ReleaseDefaultResources();
	ReleaseSysMemResources();

	return true;
}

//tB^p[^̃Zbg
bool GpuFilter::SetParams(int *track, int *check)
{
	//̃`FbN
	if(!m_pEffect) return false;

	//gbNo[
	for(unsigned int i=0; i<m_hTrack.size(); ++i){
		ProcTrack(&m_hTrack[i], track[i]);
	}

	//`FbN{bNX
	for(unsigned int i=0; i<m_hCheck.size(); ++i){
		ProcCheck(&m_hCheck[i], check[i]);
	}
	
	return true;
}

//t[Kvȏs
bool GpuFilter::SetFrame(int cur, int num)
{
	HRESULT hr;

	//̃`FbN
	if(!m_pEffect) return false;

	//ɕۑ
	m_frame = cur;
	m_frame_num = num;

	//t[ԍȂǂ̃Zbg
	if(m_hFrame) m_pEffect->SetInt(m_hFrame, m_frame);
	if(m_hFrameNum) m_pEffect->SetInt(m_hFrameNum, m_frame_num);

	//IȎQƃeNX`݂ȂƂ͂ŏI
	if(m_pTexSrc.empty()) return true;

	//eNX`̃Zbg
	std::list<int> needs_update;
	std::swap(m_mTexSrc, m_mTexSrcUpd);
	//̎_ m_mTexSrcUpd ɂׂĈڂĂ͂

	//݂Ăf[^̂܂܍ėpłꍇ̏
	std::vector<int>::const_iterator ct0;
	for(ct0 = m_nRefOffs.begin(); ct0 != m_nRefOffs.end(); ++ct0){
		stdext::hash_map<int, IDirect3DTexture9*>::iterator itr;
		itr = m_mTexSrcUpd.find(*ct0 + m_frame);

		if(itr == m_mTexSrcUpd.end()){
			//f[^݂͑Ȃ
			needs_update.push_back(*ct0 + m_frame);
		}else{
			//ėp\Ȃ̂ŃeNX`Zbg
			std::vector<TxSrcHandle>::const_iterator ct1;
			for(ct1 = m_hTxSrc.begin(); ct1 != m_hTxSrc.end(); ++ct1){
				//ItZbgȃeNX`ꍇl
				if(ct1->offset == *ct0){
					hr = m_pEffect->SetTexture(ct1->handle, itr->second);
					if(FAILED(hr)){
						ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
							TEXT("eNX`̃ZbgɎs܂"));
						return false;
					}
				}
			}
			//}bvԂœւ
			m_mTexSrc.insert(*itr);
			m_mTexSrcUpd.erase(itr);
		}
	}

	//f[^Rs[Kvꍇ̏
	std::list<int>::const_iterator ct2;
	stdext::hash_map<int, IDirect3DTexture9*> TxSrcTmp;
	stdext::hash_map<int, IDirect3DTexture9*>::iterator itt = m_mTexSrcUpd.begin();

	for(ct2 = needs_update.begin(); ct2 != needs_update.end(); ++ct2){
		if(itt == m_mTexSrcUpd.end()){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, E_OUTOFMEMORY,
				TEXT("Zbg\ȃeNX`cĂ܂"));
			return false;
		}else{
			//eNX`̃Zbg
			std::vector<TxSrcHandle>::const_iterator ct1;
			for(ct1 = m_hTxSrc.begin(); ct1 != m_hTxSrc.end(); ++ct1){
				if((m_frame + ct1->offset) == *ct2){
					hr = m_pEffect->SetTexture(ct1->handle, itt->second);
					if(FAILED(hr)){
						ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
							TEXT("eNX`̃ZbgɎs܂"));
						return false;
					}
				}
			}

			//t[ԍ̃Zbg
			TxSrcTmp.insert( Tex_Pair(*ct2, itt->second) );
			++itt;
		}
	}
	TxSrcTmp.swap(m_mTexSrcUpd);

	return true;
}

//PIXEL_YCf[^foCXɓ]
bool GpuFilter::PutImage(const PIXEL_YC *data, int n, EXEC_MT func_mt)
{
	HRESULT hr = D3D_OK;

	//ǂݗ\̂߂̃f[^XV
	m_frame_set[1] = m_frame_set[0];
	m_frame_set[0] = n;

	//eNX`I
	if((n == m_frame) && !m_TxCur){
		//݂̃t[QƃeNX`Ƃđ݂ȂꍇiTxTmp/TxRT0̎QƁj
		if(m_hTxTmp == NULL && m_hTxRT.empty()){
			//݂̃t[͎QƂȂ
			return true;
		}else{
			//݂̃t[f[^VXeT[tFXɃRs[i]͌Łj
			hr = CopyPixelData(m_pTexSys, data, func_mt);
		}
	}else{
		//ʏ̎QƃeNX`
		stdext::hash_map<int, IDirect3DTexture9*>::iterator itr;
		if(m_mTexSrc.find(n) == m_mTexSrc.end()){
			//f[^݂͑Ȃ̂œ]Kv
			itr = m_mTexSrcUpd.find(n);
			if(itr == m_mTexSrcUpd.end()){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, E_OUTOFMEMORY,
					TEXT("]̃eNX`擾ł܂ł"));
				return false;
			}
			
			//eNX`̃`FbN
			if(NULL == itr->second){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, D3DERR_INVALIDCALL,
					TEXT("eNX`쐬Ă܂"));
				return false;
			}

			//_Ci~bNeNX`͒ړ]
			hr = PutPixelData(itr->second, data, func_mt);

			//ǂɂ炸ړ
			m_mTexSrc.insert(*itr);
			m_mTexSrcUpd.erase(itr);
		}else{
			//f[^ɑ݂邽߉Ȃ
			return true;
		}
	}
	
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("t[̃Rs[Ɏs܂"));
		return false;
	}

	return true;
}

//PIXEL_YCf[^VXeRs[
bool GpuFilter::GetImage(PIXEL_YC *data, EXEC_MT func_mt)
{
	HRESULT hr;

	//񓮍̏ꍇ͏I҂
	if(m_tproc.isRunning()){
		m_tproc.Wait();
	}

	//VXeϊăRs[
	hr = CopyPixelData(data, m_pTexSys, func_mt);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr,
			TEXT("σf[^Rs[ł܂ł"));
		return false;
	}

	return true;
}

//f[^̏si/񓯊j
bool GpuFilter::ProcImage(int *next)
{
	if(!m_pEffect || !m_pd3dDevice){
		return false;
	}

	if(m_tproc.isRunning()){
		//ǂ݃t[̗\
		if(m_frame_set[0] == m_frame_set[1]){
			*next = -1;
		}else{
			*next = m_frame_set[0] + (m_frame_set[0] - m_frame_set[1]);
		}

		//񓯊ŏ
		return m_tproc.Start() ? false : true;
	}else{
		*next = -1;
		return FAILED(ProcImage()) ? false : true;
	}
}

//ɎĂ摜NA
bool GpuFilter::ClearRefFrames()
{
	stdext::hash_map<int, IDirect3DTexture9*> TxSrcTmp;
	stdext::hash_map<int, IDirect3DTexture9*>::iterator itt = m_mTexSrc.begin();

	while(itt != m_mTexSrc.end()){
		//t[ԍ肦ȂlɃZbgȂ
		TxSrcTmp.insert( Tex_Pair(-1024 - itt->first, itt->second) );
		++itt;
	}
	TxSrcTmp.swap(m_mTexSrc);

	m_frame_set[0] = m_frame_set[1] = -1;

	return true;
}

//---------------------------------------------------------------------
// private functions
//---------------------------------------------------------------------

//foCX̃T|[g󋵂̃`FbN
HRESULT GpuFilter::CheckDeviceCap()
{
	HRESULT hr;

	//|C^̃`FbN
	if(!m_pd3dDevice || !m_pD3D) return D3DERR_INVALIDCALL;

	//D3DCaps̃`FbN
	hr = m_pd3dDevice->GetDeviceCaps(&m_caps);
	if(D3D_OK != hr) return hr;

	//eNX`֘ÃeXg
	if(m_caps.TextureCaps & D3DPTEXTURECAPS_SQUAREONLY){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("񐳕eNX`𗘗pł܂"));
		return D3DERR_NOTAVAILABLE;
	}
	if(!(m_caps.TextureCaps & D3DPTEXTURECAPS_POW2)){
		; //2̗ݏȊÕTCỸeNX`
	}else if(!(m_caps.TextureCaps & D3DPTEXTURECAPS_NONPOW2CONDITIONAL)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("2̗ݏTCYȊÕeNX`͗pł܂"));
		return D3DERR_NOTAVAILABLE;
	}

    //RTtH[}bg̃`FbN
    D3DDISPLAYMODE DisplayMode;
    hr = m_pd3dDevice->GetDisplayMode( 0, &DisplayMode );
	if(D3D_OK != hr) return hr;

	bool found = false;
    if(SUCCEEDED(CheckRtTexFormat(
		m_caps.AdapterOrdinal, m_caps.DeviceType, DisplayMode.Format, D3DFMT_A16B16G16R16F)))
	{
		found = true;
		m_pixFormatAvailable[F16] = true;
	}else{
#ifdef _DEBUG
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A16B16G16R16F pł܂"));
#endif
		m_pixFormatAvailable[F16] = false;
	}

	if(SUCCEEDED(CheckRtTexFormat(
		m_caps.AdapterOrdinal, m_caps.DeviceType, DisplayMode.Format, D3DFMT_A32B32G32R32F)))
	{
		found = true;
		m_pixFormatAvailable[F32] = true;
	}else{
#ifdef _DEBUG
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A32B32G32R32F pł܂"));
#endif
		m_pixFormatAvailable[F32] = false;
	}

	if(SUCCEEDED(CheckRtTexFormat(
		m_caps.AdapterOrdinal, m_caps.DeviceType, DisplayMode.Format, D3DFMT_A16B16G16R16)))
	{
		found = true;
		m_pixFormatAvailable[U16] = true;
		m_pixFormatAvailable[B16] = true;
	}else{
#ifdef _DEBUG
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A16B16G16R16 pł܂"));
#endif
		m_pixFormatAvailable[U16] = false;
		m_pixFormatAvailable[B16] = false;
	}

	if(SUCCEEDED(CheckRtTexFormat(
		m_caps.AdapterOrdinal, m_caps.DeviceType, DisplayMode.Format, D3DFMT_Q16W16V16U16)))
	{
		found = true;
		m_pixFormatAvailable[S16] = true;
	}else{
#ifdef _DEBUG
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_Q16W16V16U16 pł܂"));
#endif
		m_pixFormatAvailable[S16] = false;
	}
	

	if(SUCCEEDED(CheckRtTexFormat(
		m_caps.AdapterOrdinal, m_caps.DeviceType, DisplayMode.Format, D3DFMT_A2B10G10R10)))
	{
		found = true;
		m_pixFormatAvailable[U10] = true;
	}else{
#ifdef _DEBUG
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("D3DFMT_A2B10G10R10 pł܂"));
#endif
		m_pixFormatAvailable[U10] = false;
	}

	if(!found){   
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("p\ȃsNZtH[}bgЂƂ܂"));
		return D3DERR_NOTAVAILABLE;
	}

	return hr;
}

//_̐錾
HRESULT GpuFilter::CreateVert()
{
	HRESULT hr;

    //foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pVert);

	PPVERT Quad[4] =
	{
		{         -0.5f,          -0.5f, 1.0f, 1.0f, 0.0f, 0.0f },
		{ m_width -0.5f,          -0.5f, 1.0f, 1.0f, 1.0f, 0.0f },
		{         -0.5f, m_height -0.5f, 1.0f, 1.0f, 0.0f, 1.0f },
		{ m_width -0.5f, m_height -0.5f, 1.0f, 1.0f, 1.0f, 1.0f }
	};

	hr = m_pd3dDevice->CreateVertexBuffer(sizeof(PPVERT) * 4,
		D3DUSAGE_WRITEONLY, 0, D3DPOOL_DEFAULT, &m_pVert, NULL );

	//obt@ɓ]
	LPVOID pVBData;
	if(SUCCEEDED(m_pVert->Lock(0, 0, &pVBData, D3DLOCK_NOSYSLOCK))){
		CopyMemory(pVBData, Quad, sizeof(Quad));
		m_pVert->Unlock();
	}

	return hr;
}

//EffectReadergăGtFNg쐬
HRESULT GpuFilter::CreateEffect()
{
	const DWORD ShaderFlags = D3DXFX_NOT_CLONEABLE|D3DXFX_LARGEADDRESSAWARE;
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(m_pEffect);

	//GtFNg̍쐬
	hr = m_pEffectReader.CreateEffect(m_pd3dDevice, ShaderFlags, &m_pEffect);
	if(FAILED(hr)) return hr;

	//̃p[^nh̎擾
	m_hTxTmp    = m_pEffect->GetParameterByName(NULL, "TxTmp");
	m_hFrame    = m_pEffect->GetParameterByName(NULL, "Frame");
	m_hFrameNum = m_pEffect->GetParameterByName(NULL, "FrameNum");
	m_hSrcSize  = m_pEffect->GetParameterByName(NULL, "SrcSize");
	m_hPixScale = m_pEffect->GetParameterByName(NULL, "PixScale");

	//c̃p[^ƃnh擾
	hr = ParseEffect();
	if(FAILED(hr)){
		SAFE_RELEASE(m_pEffect);
		return hr;
	}

	//tH[}bg̃eXg
	hr = TestEffect();

	return hr;
}

//GtFNgǂݎĕKvȃ\[X𒲂ׂ
HRESULT GpuFilter::ParseEffect()
{
	CHAR buf[32];
	
	if(!m_pEffect) return D3DERR_INVALIDCALL;

	//όȐݒ̓ǂݏo
	bool tech_sel = false; //eNjbNI1̂

	//gbN
	m_hTrack.clear();
	const std::vector<TrackInfo> &ti = m_pEffectReader.GetTrackInfo();
	for(unsigned int i=0; i<ti.size(); ++i){
		TrackHandle ht = { NULL, ti[i].max, ti[i].min, DEF, 1 };

		StringCchPrintfA(buf, ARRAYSIZE(buf), "Track%d", i);
		ht.handle = m_pEffect->GetParameterByName(NULL, buf);
		if(ht.handle == NULL){
			break;
		}

		//usagẽ`FbN
		if(ti[i].usage == "normalize"){
			ht.usage = NORM;
			ht.abs_max = max(abs(ht.min), abs(ht.max));

			//0Z
			if(ht.abs_max == 0){
				ht.abs_max = 1;
			}
		}else if(ti[i].usage == "technique"){
			if(tech_sel){
				//Iς
				ht.usage = NONE;
			}else{
				ht.usage = TECH;
				tech_sel = true;
			}
		}else if(ti[i].usage == "even"){
			ht.usage = EVEN;
		}else if(ti[i].usage == "odd"){
			ht.usage = ODD;
		}else if(ti[i].usage == "odd0"){
			ht.usage = ODD0;
		}
		
		m_hTrack.push_back(ht);
	}
	
	//`FbN{bNX
	m_hCheck.clear();
	const std::vector<CheckInfo> &ci = m_pEffectReader.GetCheckInfo();
	for(unsigned int i=0; i<ci.size(); ++i){
		CheckHandle hc = { NULL, DEF };

		StringCchPrintfA(buf, ARRAYSIZE(buf), "Check%d", i);
		hc.handle = m_pEffect->GetParameterByName(NULL, buf);
		if(hc.handle == NULL){
			break;
		}

		//usagẽ`FbN
		if(ci[i].usage == "normalize"){
			hc.usage = NORM;
		}else if(ci[i].usage == "technique"){
			if(tech_sel){
				//Iς
				hc.usage = NONE;
			}else{
				hc.usage = TECH;
				tech_sel = true;
			}
		}else if(ci[i].usage == "even"){
			hc.usage = NORM;
		}else if(ci[i].usage == "odd" || ci[i].usage == "odd0"){
			hc.usage = ODD; //`FbN̏ꍇ͓
		}

		m_hCheck.push_back(hc);
	}

	//J[l
	m_hKern.clear();
	for(int i=0; i<KERN_MAX; ++i){
		KernelHandle hk = { NULL, NULL, 0 };
		D3DXPARAMETER_DESC desc;

		StringCchPrintfA(buf, ARRAYSIZE(buf), "PixelKernel%d", i);
		hk.pix = m_pEffect->GetParameterByName(NULL, buf);
		StringCchPrintfA(buf, ARRAYSIZE(buf), "TexelKernel%d", i);
		hk.tex = m_pEffect->GetParameterByName(NULL, buf);
		if(hk.pix == NULL || hk.tex == NULL){
			break;
		}

		if(SUCCEEDED(m_pEffect->GetParameterDesc(hk.pix, &desc))){
			hk.size = desc.Elements;
		}
		m_hKern.push_back(hk);
	}

	//\[X摜eNX`
	std::list<int> offs;
	m_hTxSrc.clear();
	for(int i=0; i<TREF_MAX; ++i){
		StringCchPrintfA(buf, ARRAYSIZE(buf), "TxSrc%d", i);
		D3DXHANDLE hh = m_pEffect->GetParameterByName(NULL, buf);
		if(hh){
			TxSrcHandle sh = { hh, 0 };
			//ItZbgԍ̎擾
			hh = m_pEffect->GetAnnotationByName(hh, "frame_offset");
			if(hh){
				m_pEffect->GetInt(hh, &sh.offset);
			}
			m_hTxSrc.push_back(sh);
			offs.push_back(sh.offset);
		}else{
			break;
		}
	}
	offs.sort();
	offs.unique();
	m_nRefOffs.clear();

	//ۂɕKvȃ\[X摜̃\[Xς
	if(m_hTxSrc.empty()){
		//Iȃ\[X摜͕sv
		m_pTexSrc.clear();
		m_TxCur = false;
	}else{
		//dt[͏
		m_pTexSrc.resize(offs.size());
		m_TxCur = false;

		//\[X摜ɃItZbg0̃t[ꍇ
		//Rs[ĎgƂɂ
		std::list<int>::iterator res = offs.begin();
		while(res != offs.end()){
			if(*res == 0){
				m_TxCur = true;
			}
			m_nRefOffs.push_back(*res);
			++res;
		}
		std::sort(m_nRefOffs.begin(), m_nRefOffs.end());
	}

	//RTeNX`
	m_hTxRT.clear();
	for(int i=0; i<TXRT_MAX; ++i){
		StringCchPrintfA(buf, ARRAYSIZE(buf), "TxRT%d", i);
		D3DXHANDLE hh = m_pEffect->GetParameterByName(NULL, buf);
		if(hh){
			m_hTxRT.push_back(hh);
		}else{
			break;
		}
	}
	//RTeNX`͍Œ1ʁiTxTmpȂ2ʁjpӂ
	m_pTexRT.resize(max(m_hTxTmp ? 2U : 1U, m_hTxRT.size()));

	//eNjbN
	m_hTech.clear();
	for(int i=0; i<TECH_MAX; ++i){
		D3DXTECHNIQUE_DESC dsc = { NULL, 0 , 0 };
		TechHandle th;

		StringCchPrintfA(buf, ARRAYSIZE(buf), "ProcessImage%d", i);
		th.handle = m_pEffect->GetTechniqueByName(buf);
		if(th.handle == NULL){
			break;
		}

		m_pEffect->GetTechniqueDesc(th.handle, &dsc);
		th.npass = dsc.Passes;
		th.rt.resize(dsc.Passes);

		//RT̎w𒲍
		bool rt_sel = false;
		for(unsigned int j=0; j<dsc.Passes; ++j){
			D3DXHANDLE hp = m_pEffect->GetPass(th.handle, j);
			for(unsigned int k=0; k<RtTbl::rt_max; ++k){
				if(hp){
					StringCchPrintfA(buf, ARRAYSIZE(buf), "RenderTarget%u", k);
					D3DXHANDLE ha = m_pEffect->GetAnnotationByName(hp, buf);

					//TxRTn  n w肷
					INT n;
					if(ha && S_OK == m_pEffect->GetInt(ha, &n)){
						th.rt[j][k] = n;
						rt_sel = true;
					}
				}
			}
		}

		//RTw肪ꍇvector
		if(!rt_sel) th.rt.clear();

		m_hTech.push_back(th);
	}


	return S_OK;
}

//GtFNgǂݎƂɊeeXgs
HRESULT GpuFilter::TestEffect()
{
	LPCSTR str;
	D3DXHANDLE tmpx;
	TCHAR msg[256];
	
	if(!m_pEffect) return D3DERR_INVALIDCALL;

	//sNZtH[}bg
	PIXEL_FORMAT tmp_fmt = F32; //ftHg
	tmpx = m_pEffect->GetParameterByName(NULL, "FilterVersion@img_format");
	if(tmpx && SUCCEEDED(m_pEffect->GetString(tmpx, &str))){
		if(strcmp(str, "S16") == 0){
			tmp_fmt = S16;
		}else if(strcmp(str, "B16") == 0){
			tmp_fmt = B16;
		}else if(strcmp(str, "F16") == 0){
			tmp_fmt = F16;
		}else if(strcmp(str, "U16") == 0){
			tmp_fmt = U16;
		}

		//vtH[}bgɑΉĂ邩
		if(tmp_fmt != F32 && !m_pixFormatAvailable[tmp_fmt]){
			StringCchPrintf(msg, ARRAYSIZE(msg),
#ifdef UNICODE
				TEXT("\"%S\"")
#else
				TEXT("\"%s\"")
#endif
				TEXT("tH[}bg͌ݗpGPUł̓T|[gĂ܂B\n")
				TEXT("\"F32\"Ƃď܂A삵Ȃ\܂B"),
				str);
			MessageBox(NULL, msg, GetFilterName(), MB_ICONWARNING|MB_SYSTEMMODAL);
			tmp_fmt = F32;
		}
	}
	m_pixFormat = m_pixFormatAvailable[tmp_fmt] ? tmp_fmt : Unknown;

	//ΉtH[}bg݂Ȃ
	if(m_pixFormat == Unknown){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, E_NOTIMPL,
			TEXT("p\ȃtH[}bg܂"));
		return E_NOTIMPL;
	}
	
	//PSo[W̃eXg
	int min_ps[2] = { 2, 0 };
	tmpx = m_pEffect->GetParameterByName(NULL, "FilterVersion@min_ps_ver");
	if(tmpx && SUCCEEDED(m_pEffect->GetIntArray(tmpx, min_ps, 2))){
		//vo[W𖞂Ȃꍇx
		if(m_caps.PixelShaderVersion < D3DPS_VERSION(min_ps[0], min_ps[1])){
			StringCchPrintf(msg, ARRAYSIZE(msg), TEXT("̃tB^̓ɂ ")
				TEXT("ps_%x_%x ȍ~T|[gGPUKvłB\n")
				TEXT("ݗpGPUł͐ɓ삵Ȃ\܂B"),
				min_ps[0], min_ps[1]);
			MessageBox(NULL, msg, GetFilterName(), MB_ICONWARNING|MB_SYSTEMMODAL);
		}
	}

	//MRT֘ÃeXg
	bool mrt_warn = true;
	std::vector<TechHandle>::const_iterator itt = m_hTech.begin();
	for( ; itt != m_hTech.end(); ++itt){
		for(unsigned int i=0; i<itt->rt.size(); ++i){
			unsigned int j;
			const RtTbl &rtt = itt->rt[i];
			for(j=0; j<rtt.rt_max; ++j){
				if(rtt[j] < 0){
					break;
				}else if(rtt[j] >= (int)m_hTxRT.size()){
					ShowMessageBox(__FILE__, (DWORD)__LINE__, E_INVALIDARG,
						TEXT("݂Ȃ_[^[Qbgo͂Ɏw肳Ă܂B"));
					return E_INVALIDARG;
				}

				if(rtt.count(rtt[j]) != 1){
					ShowMessageBox(__FILE__, (DWORD)__LINE__, E_INVALIDARG,
						TEXT("_[^[Qbgɏo͐ƂĎw肳Ă܂B"));
					return E_INVALIDARG;
				}
			}
			
			//RTwsĂeNjbNł͑S̃pXŏo͎w肪Kv
			if(rtt.count(-1) == rtt.rt_max){
				ShowMessageBox(__FILE__, (DWORD)__LINE__, E_INVALIDARG,
					TEXT("_[^[Qbgw𗘗pĂeNjbNł́A\n")
					TEXT("ׂẴpXŏo͐w肵ĂKv܂B"));
				return E_INVALIDARG;
			}

			//MRT̃n[hEFAT|[geXg
			if(mrt_warn && j >= m_caps.NumSimultaneousRTs){
				mrt_warn = false;
				MessageBox(NULL, TEXT("ݗpGPUł͓ɏo͂ł^[Qbg")
					TEXT("sĂ܂B\nȏo͂Ȃ\܂B"),
					GetFilterName(), MB_ICONWARNING|MB_SYSTEMMODAL);
			}
		}
	}

	//Qƃt[(TxSrc#)Ȃꍇ̏
	if((m_hTxTmp || !m_hTxRT.empty()) && m_nRefOffs.empty()){
		//TxTmp or TxRT0 ɏf[^i[Kv
		m_nRefOffs.push_back(0);

		//RTЂƂTxRT0݂ꍇ͌x
		if(m_pTexRT.size() == 1 && m_hTxRT.size() == 1){
			MessageBox(NULL,
				TEXT("\"TxSrc#\"\"TxTmp\"݂\"TxRT0\"݂̂LłB\n")
				TEXT("\"TxRT0\"QƂďo͂Ǝg邱ƂɂȂA\n")
				TEXT("Ȍʂ͕ۏ؂܂B"),
				GetFilterName(), MB_ICONWARNING|MB_SYSTEMMODAL);
		}
	}

	return S_OK;
}

//VXeɕێ郊\[X쐬
HRESULT GpuFilter::CreateSysMemResources()
{
	HRESULT hr;

	if(!m_pd3dDevice || !m_pEffect) return D3DERR_INVALIDCALL;

	//_錾̍쐬
    hr = m_pd3dDevice->CreateVertexDeclaration(PPVERT::Decl, &m_pVertDec);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_錾̍쐬Ɏs܂"));
		return false;
	}

	//VXẽeNX`
	hr = CreateTexture(&m_pTexSys, 0, D3DPOOL_SYSTEMMEM);
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("VXeeNX`̍쐬Ɏs܂"));
		return hr;
	}

	return hr;
}

//GPUɑ݂郊\[X쐬FXgɂ͕AKv
HRESULT GpuFilter::CreateDefaultResources()
{
	HRESULT hr;

	//_
	hr = CreateVert();
	if(FAILED(hr)){
		ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("_̃ZbgɎs܂"));
		return hr;
	}

	//QƃeNX`
	m_mTexSrc.clear();
	m_mTexSrcUpd.clear();
	for(unsigned int i=0; i<m_pTexSrc.size(); ++i){
		hr = CreateTexture(&m_pTexSrc[i], D3DUSAGE_DYNAMIC, D3DPOOL_DEFAULT);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("eNX`̍쐬Ɏs܂"));
			return hr;
		}
		//lƂĂ肦Ȃl
		m_mTexSrc.insert( Tex_Pair((-1024 - (int)i), m_pTexSrc[i]) );
	}

	//RTeNX`
	for(unsigned int i=0; i<m_pTexRT.size(); ++i){
		hr = CreateTexture(&m_pTexRT[i], D3DUSAGE_RENDERTARGET, D3DPOOL_DEFAULT);
		if(FAILED(hr)){
			ShowMessageBox(__FILE__, (DWORD)__LINE__, hr, TEXT("RTeNX`̍쐬Ɏs܂"));
			return hr;
		}
	}
	
	return hr;
}

//GtFNgɌŒlݒ肷
HRESULT GpuFilter::SetEffectResources()
{
	//foCX̃`FbN
	if(!m_pd3dDevice || !m_pEffect) return D3DERR_INVALIDCALL;

	//J[l̐ݒ
	float rx = 1.0f/m_width;
	float ry = 1.0f/m_height;
	for(unsigned int i=0; i<m_hKern.size(); ++i){
		D3DXVECTOR4 *tmp = new(std::nothrow) D3DXVECTOR4[m_hKern[i].size];
		if(tmp){
			m_pEffect->GetVectorArray(m_hKern[i].pix, tmp, m_hKern[i].size);

			for(int j=0; j<m_hKern[i].size; ++j){
				tmp[j].x *= rx;
				tmp[j].y *= ry;
			}

			m_pEffect->SetVectorArray(m_hKern[i].tex, tmp, m_hKern[i].size);
			delete[] tmp;
		}
	}

	//摜TCY
	if(m_hSrcSize){
		D3DXVECTOR4 size((float)m_width, (float)m_height, 0.0f, 0.0f);
		m_pEffect->SetVector(m_hSrcSize, &size);
	}

	//sNZl̃XP[O
	if(m_hPixScale){
		static const D3DXVECTOR4 scale[Unknown + 1] = {
			D3DXVECTOR4( 4096.0f,      0.0f, 0.0f, 0.0f), //F32
			D3DXVECTOR4( 4096.0f,      0.0f, 0.0f, 0.0f), //F16
			D3DXVECTOR4(65535.0f,      0.0f, 0.0f, 0.0f), //U16
			D3DXVECTOR4(65535.0f, -32768.0f, 0.0f, 0.0f), //B16
			D3DXVECTOR4(32767.0f,      0.0f, 0.0f, 0.0f), //S16
			D3DXVECTOR4(    0.0f,      0.0f, 0.0f, 0.0f), //U10
			D3DXVECTOR4(    0.0f,      0.0f, 0.0f, 0.0f)  //Unknown
		};
		m_pEffect->SetVector(m_hPixScale, &scale[m_pixFormat]);
	}

	//RTeNX`
	if(!m_hTxRT.empty() && (m_hTxRT.size() == m_pTexRT.size())){
		for(unsigned int i=0; i<m_hTxRT.size(); ++i){
			m_pEffect->SetTexture(m_hTxRT[i], m_pTexRT[i]);
		}
	}

	return D3D_OK;
}

//VXeɕێ郊\[XJ
HRESULT GpuFilter::ReleaseSysMemResources()
{
	SAFE_RELEASE(m_pTexSys);
	SAFE_RELEASE(m_pVertDec);
	return D3D_OK;
}

//GPUɑ݂郊\[XJFXgɗp
HRESULT GpuFilter::ReleaseDefaultResources()
{
	SAFE_RELEASE(m_pVert);
	for(unsigned int i=0; i<m_pTexSrc.size(); ++i) SAFE_RELEASE(m_pTexSrc[i]);
	for(unsigned int i=0; i<m_pTexRT.size(); ++i) SAFE_RELEASE(m_pTexRT[i]);
	return D3D_OK;
}

//foCXXgɃZbgs
HRESULT GpuFilter::Reset()
{
	HRESULT hr;

	//foCX̃`FbN
	if(!m_pd3dDevice || !m_pEffect) return D3DERR_INVALIDCALL;

	//GPU\[X̊J
	ReleaseDefaultResources();
	m_pEffect->OnLostDevice();

	//Zbg
	D3DPRESENT_PARAMETERS dpp;

	ZeroMemory(&dpp, sizeof(dpp));
	dpp.Windowed = TRUE;
	dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	dpp.BackBufferFormat = D3DFMT_UNKNOWN;

	hr = m_pd3dDevice->Reset(&dpp);
	if(FAILED(hr)){
		return hr;
	}

	hr = m_pEffect->OnResetDevice();
	if(FAILED(hr)){
		return hr;
	}

	//\[X̍Ċmۂ͍sȂ
	return hr;
}

//摜̏{
HRESULT GpuFilter::ProcImage()
{
	HRESULT hr = E_FAIL;

	if(m_pEffect == NULL || m_pd3dDevice == NULL){
		return hr;
	}

	//eNjbN݂Ȃ
	if(m_hTech.empty()){
		return hr;
	}

	//KvȂ猻݂̉摜pӂ
	if(m_hTxTmp || !m_hTxRT.empty()){
		//ItZbg0̎QƃeNX`ꍇɂTxTmpɃZbg
		//TxRT0̒g͕sƂ
		if(m_TxCur && m_hTxTmp){
			stdext::hash_map<int, IDirect3DTexture9*>::const_iterator ct;
			ct = m_mTexSrc.find(m_frame);
			if(ct != m_mTexSrc.end()){
				hr = m_pEffect->SetTexture(m_hTxTmp, ct->second);
			}
		}else if(!m_TxCur){
			//VXeeNX`ɃRs[ς݂Ȃ̂ł]
			hr = PutPixelData(m_pTexRT[0], m_pTexSys);
			if(m_hTxTmp){
				//Jn_łm_hTxTmpRT0Ɠ̂Ƃ݂Ȃ
				m_pEffect->SetTexture(m_hTxTmp, m_pTexRT[0]);
			}
		}else{
			//ItZbg0̎QƃeNX`
			//TxTmp݂͑TxRT0͑݁iTxRT0̒g͕sj
			hr = S_OK;
		}
		
		if(FAILED(hr)){
			return hr;
		}	
	}

	//_OJn
	if(SUCCEEDED(m_pd3dDevice->BeginScene()))
	{
		//RTۑ
		IDirect3DSurface9 *rtOld = NULL;
		m_pd3dDevice->GetRenderTarget(0, &rtOld);

		//eNjbNw
		m_pEffect->SetTechnique(m_hTech[m_tech].handle);

		// Set the vertex declaration
		m_pd3dDevice->SetVertexDeclaration(m_pVertDec);

		// Draw the quad
		UINT cPasses;
		m_pEffect->Begin( &cPasses, 0 );

		//Pass
		for(UINT p=0; p<cPasses; ++p){
			//RenderTarget̐ݒ
			if(!m_hTech[m_tech].rt.empty()){
				//MRTw
				SetRenderTarget(m_hTech[m_tech].rt[p]);
			}else if(m_pTexRT.size() > 1){
				//MRTw肪ꍇ1,0Ɍ݂ɏói]݊j
				SetRenderTarget(m_pTexRT[(p + 1U) & 1U]);
			}else{
				//TxTmpTxRT1ȏオ݂ȂƂRT͂ЂƂ
				SetRenderTarget(m_pTexRT[0]);
			}

			//Render
			m_pEffect->BeginPass(p);
			m_pd3dDevice->SetStreamSource(0, m_pVert, 0, sizeof(PPVERT));
			m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
			m_pEffect->EndPass();

			//o͌ʃeNX`̐ݒ
			if(m_hTxTmp){
				m_pEffect->SetTexture(m_hTxTmp, m_pTexRT[(p + 1U) & 1U]);
			}
		}

		m_pEffect->End();

		//End scene
		m_pd3dDevice->EndScene();
		
		//ŏIo͌ʂ߂i݂0ԖڂRTj
		GetRenderTargetData();

		//RTɖ߂
		if(rtOld){
			m_pd3dDevice->SetRenderTarget(0, rtOld);
			rtOld->Release();
		}
		for(DWORD i=1; i<m_caps.NumSimultaneousRTs; ++i){
			m_pd3dDevice->SetRenderTarget(i, NULL);
		}
	}

	return true;
}

//eNX`쐬
HRESULT GpuFilter::CreateTexture(IDirect3DTexture9 **ppTex, DWORD usage, D3DPOOL pool)
{
	D3DFORMAT fmt = PicFmtToD3DFmt(m_pixFormat);

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;
	
	//Ŝ
	SAFE_RELEASE(*ppTex);

	//eNX`쐬
	return m_pd3dDevice->CreateTexture(m_width, m_height, 1, usage, fmt, pool, ppTex, NULL);
}

//T[tFX쐬
HRESULT GpuFilter::CreateSurface(IDirect3DSurface9 **ppSrf, DWORD usage, D3DPOOL pool)
{
	D3DFORMAT fmt = PicFmtToD3DFmt(m_pixFormat);

	//foCX̃`FbN
	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//Ŝ
	SAFE_RELEASE(*ppSrf);

	if(usage & D3DUSAGE_RENDERTARGET){
		//_[^[Qbg쐬
		return m_pd3dDevice->CreateRenderTarget(
			m_width, m_height, fmt, D3DMULTISAMPLE_NONE, 0, FALSE, ppSrf, NULL);
	}else{
		//ItXN[T[tFX쐬
		return m_pd3dDevice->CreateOffscreenPlainSurface(m_width, m_height, fmt, pool, ppSrf, NULL);
	}
}

//VXeeNX`PIXEL_YC]
HRESULT GpuFilter::CopyPixelData(IDirect3DTexture9 *sys, const PIXEL_YC *src, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!sys || !m_pCpCvt) return D3DERR_INVALIDCALL;

	//VXẽT[tFXbN
	hr = sys->LockRect(0, &lRect, NULL, D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̃Rs[
	CpCvtParam para =
	{ lRect.pBits, const_cast<PIXEL_YC*>(src), m_height, m_width, lRect.Pitch, m_wMax };
	
	if(func_mt){
		func_mt(m_pCpCvt->GetYCtoX(), &para, NULL);
	}else{
		m_pCpCvt->ProcYCtoX(para);
	}
	
	return sys->UnlockRect(0);
}

//DynamicTexture𗘗pړ]
HRESULT GpuFilter::PutPixelData(IDirect3DTexture9 *dst, const PIXEL_YC *src, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!dst || !m_pCpCvt) return D3DERR_INVALIDCALL;

	//_Ci~bNeNX`̃T[tFXbN
	hr = dst->LockRect(0, &lRect, NULL, D3DLOCK_DISCARD|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̓]
	CpCvtParam para =
	{ lRect.pBits, const_cast<PIXEL_YC*>(src), m_height, m_width, lRect.Pitch, m_wMax };
	
	if(func_mt){
		func_mt(m_pCpCvt->GetYCtoX(), &para, NULL);
	}else{
		m_pCpCvt->ProcYCtoX(para);
	}
	
	return dst->UnlockRect(0);
}

//VXeeNX`oRԐړ]
HRESULT GpuFilter::PutPixelData(IDirect3DTexture9 *dst, IDirect3DTexture9 *sys)
{
	HRESULT hr;

	if(!m_pd3dDevice) return D3DERR_INVALIDCALL;

	//eNX`]
	hr = m_pd3dDevice->UpdateTexture(sys, dst);

	return hr;
}

//RTeNX`RTƂăZbg
HRESULT GpuFilter::SetRenderTarget(IDirect3DTexture9 *rtt)
{
	HRESULT hr;
	IDirect3DSurface9 *rt = NULL;
	
	if(!m_pd3dDevice || !rtt){
		return D3DERR_INVALIDCALL;
	}

	//RTeNX`T[tFX̎擾
	hr = rtt->GetSurfaceLevel(0, &rt);
	if(FAILED(hr)){
		return hr;
	}

	//RenderTarget̐ݒ
	hr = m_pd3dDevice->SetRenderTarget(0, rt);
	SAFE_RELEASE(rt);

	return hr;
}

//RTeNX`RTƂăZbg(MRTΉ)
HRESULT GpuFilter::SetRenderTarget(const RtTbl &rt)
{
	HRESULT hr = S_OK;
	
	if(!m_pd3dDevice){
		return D3DERR_INVALIDCALL;
	}

	//RT0𖳌邱Ƃ͂łȂ
	if(rt[0] < 0){
		return D3DERR_INVALIDCALL;
	}

	for(unsigned int i=0; i<max(rt.rt_max, m_caps.NumSimultaneousRTs); ++i){
		if(rt[i] < 0){
			//RT𖳌
			hr = m_pd3dDevice->SetRenderTarget(i, NULL);
			if(FAILED(hr)){
				break;
			}else{
				continue;
			}
		}else if(rt[i] >= (int)m_pTexRT.size()){
			//RTw肪s
			return D3DERR_INVALIDCALL;
		}

		//RTeNX`T[tFX̎擾
		IDirect3DSurface9 *srf = NULL;
		hr = m_pTexRT[rt[i]]->GetSurfaceLevel(0, &srf);
		if(FAILED(hr)){
			break;
		}

		//RenderTarget̐ݒ
		hr = m_pd3dDevice->SetRenderTarget(i, srf);
		SAFE_RELEASE(srf);
		if(FAILED(hr)){
			break;
		}
	}

	return hr;
}

//RTf[^VXeɃRs[
HRESULT GpuFilter::GetRenderTargetData()
{
	HRESULT hr;
	IDirect3DSurface9 *src = NULL;
	IDirect3DSurface9 *dst = NULL;

	if(!m_pd3dDevice || !m_pTexSys){
		return D3DERR_INVALIDCALL;
	}

	//݂RTT[tFX̎擾i0Ԗڂ擾j
	hr = m_pd3dDevice->GetRenderTarget(0, &src);
	if(FAILED(hr)){
		return hr;
	}
	//]
	hr = m_pTexSys->GetSurfaceLevel(0, &dst);
	if(FAILED(hr)){
		SAFE_RELEASE(src);
		return hr;
	}

	//VXeɓ]
	hr = m_pd3dDevice->GetRenderTargetData(src, dst);
	SAFE_RELEASE(src);
	SAFE_RELEASE(dst);

	return hr;
}

//VXeeNX`猳PIXEL_YCƂē]
HRESULT GpuFilter::CopyPixelData(PIXEL_YC *dst, IDirect3DTexture9 *sys, EXEC_MT func_mt)
{
	HRESULT hr;
	D3DLOCKED_RECT lRect;

	if(!sys || !m_pCpCvt) return D3DERR_INVALIDCALL;

	//VXẽeNX`bN
	hr = sys->LockRect(0, &lRect, NULL, D3DLOCK_READONLY|D3DLOCK_NOSYSLOCK);
	if(hr != D3D_OK) return hr;

	//f[^̃Rs[
	CpCvtParam para = { lRect.pBits, dst, m_height, m_width, lRect.Pitch, m_wMax };
	
	if(func_mt){
		func_mt(m_pCpCvt->GetXtoYC(), &para, NULL);
	}else{
		m_pCpCvt->ProcXtoYC(para);
	}

	return sys->UnlockRect(0);
}

//G[̕\
HRESULT GpuFilter::ShowMessageBox(const char *file, DWORD line, HRESULT hr, const TCHAR *msg)
{
#ifdef _DEBUG
	return DXTrace(file, line, hr, msg, TRUE);
#else
	MessageBox(NULL, msg, GetFilterName(), MB_ICONERROR|MB_SYSTEMMODAL);
	return S_OK;
#endif
}

//gbNo[̒l
HRESULT GpuFilter::ProcTrack(const TrackHandle *h, int track)
{
	int n;
	float f;

	if(!h || !m_pEffect) return E_FAIL;

	//p[^
	if(h->handle == NULL || h->usage == NONE){
		return S_OK;
	}

	//p@ʂɏ
	switch(h->usage){
		case NORM:
		case NORM2:
			if(h->abs_max != 0){
				f = track/(float)(h->abs_max);
			}else{
				f = 0.0f;
			}
			return m_pEffect->SetFloat(h->handle, f);
		case TECH: //eNjbNw
			n = (track < 0) ? 0 : track;
			n = (n >= (int)m_hTech.size()) ? ((int)m_hTech.size() - 1) : track;
			m_tech = n;
			return m_pEffect->SetInt(h->handle, track);
		case EVEN:
			return m_pEffect->SetInt(h->handle, track * 2);
		case ODD:
			return m_pEffect->SetInt(h->handle, track * 2 + 1);
		case ODD0:
			n = (track > 0) ? (track * 2 - 1) : (track * 2 + 1);
			return m_pEffect->SetInt(h->handle, (track == 0) ? 0 : n);
		case NONE:
		case DEF:
		default:
			return m_pEffect->SetInt(h->handle, track);
	}
}

//`FbN{bNX̏
HRESULT GpuFilter::ProcCheck(const CheckHandle *h, int check)
{
	int n;

	if(!h || !m_pEffect) return E_FAIL;

	//p[^
	if(h->handle == NULL || h->usage == NONE){
		return S_OK;
	}

	//p@ʂɏ
	switch(h->usage){
		case NORM: //[0,1]Ƀm[}CY
			return m_pEffect->SetFloat(h->handle, (check ? 1.0f : 0.0f));
		case NORM2: //[-1,1]Ƀm[}CY
			return m_pEffect->SetFloat(h->handle, (check ? 1.0f : -1.0f));
		case TECH: //eNjbNw
			n = (check < 0) ? 0 : check;
			n = (n >= (int)m_hTech.size()) ? ((int)m_hTech.size() - 1) : check;
			m_tech = n;
			return m_pEffect->SetBool(h->handle, check);
		case EVEN:
			return m_pEffect->SetInt(h->handle, check * 2);
		case ODD:
		case ODD0:
			return m_pEffect->SetInt(h->handle, check * 2 - 1);
		case NONE:
		case DEF:
		default:
			return m_pEffect->SetBool(h->handle, check);
	}
}

//tH[}bgp\ǂ̃eXg
HRESULT GpuFilter::CheckRtTexFormat(UINT ad, D3DDEVTYPE dt, D3DFORMAT af, D3DFORMAT tgt) const
{
	if(!m_pD3D){
		return D3DERR_INVALIDCALL;
	}
	
	HRESULT hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, D3DUSAGE_RENDERTARGET, D3DRTYPE_TEXTURE, tgt);
	if(S_OK != hr) return hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, D3DUSAGE_DYNAMIC, D3DRTYPE_TEXTURE, tgt);
	if(S_OK != hr) return hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, 0, D3DRTYPE_TEXTURE, tgt);
#if 0 //݂̓eNX`̂
	if(S_OK != hr) return hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, D3DUSAGE_RENDERTARGET, D3DRTYPE_SURFACE, tgt);
	if(S_OK != hr) return hr;
	hr = m_pD3D->CheckDeviceFormat(ad, dt, af, 0, D3DRTYPE_SURFACE, tgt);
#endif
	return hr;
}

//eNX`tH[}bg̕\
const TCHAR *GpuFilter::GetPixelFormatStr() const
{
	switch(m_pixFormat){
		case F32:
			return TEXT("F32");
		case F16:
			return TEXT("F16");
		case U16:
			return TEXT("U16");
		case S16:
			return TEXT("S16");
		case B16:
			return TEXT("B16");
		case U10:
			return TEXT("U10");
		case Unknown:
		default:
			return TEXT("Unknown");
	}
}

//PIXEL_FORMATD3DFORMATɕϊ
D3DFORMAT GpuFilter::PicFmtToD3DFmt(PIXEL_FORMAT p) const
{
	switch(p){
		case F16:
			return D3DFMT_A16B16G16R16F;
		case U16:
		case B16:
			return D3DFMT_A16B16G16R16;
		case S16:
			return D3DFMT_Q16W16V16U16;
		case U10:
			return D3DFMT_A2B10G10R10;
		case F32:
		case Unknown:
		default:
			return D3DFMT_A32B32G32R32F;
	}
}

//PIXEL_FORMATCpCvtFmtɕϊ
CpCvt::CpCvtFmt GpuFilter::PicFmtToCpCvtFmt(PIXEL_FORMAT p) const
{
	switch(p){
		case F32:
			return CpCvt::FLOAT32;
		case F16:
			return CpCvt::FLOAT16;
		case B16:
			return CpCvt::BIASED16;
		case U16:
		case S16:
			return CpCvt::INT16;
		case U10:
		case Unknown:
		default:
			return CpCvt::UNKNOWN;
	}
}

//tB^
const TCHAR *GpuFilter::GetFilterName() const
{
	const TCHAR *name = m_pEffectReader.GetFilterNameT();
	return (name != NULL) ? name : TEXT("GPUJX^"); 
}

//ʃXbhŏ
void __cdecl GpuFilter::AsyncProc(void *param)
{
	GpuFilter *pf = static_cast<GpuFilter*>(param);
	
	if(pf){
		pf->ProcImage();
	}
}
